# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from typing import Optional

from libcst.codemod._cli import _calculate_module
from libcst.testing.utils import UnitTest, data_provider


class TestPackageCalculation(UnitTest):
    @data_provider(
        (
            # Providing no root should give back no module.
            (None, "/some/dummy/file.py", None),
            # Providing a file outside the root should give back no module.
            ("/home/username/root", "/some/dummy/file.py", None),
            ("/home/username/root/", "/some/dummy/file.py", None),
            ("/home/username/root", "/home/username/file.py", None),
            # Various files inside the root should give back valid modules.
            ("/home/username/root", "/home/username/root/file.py", "file"),
            ("/home/username/root/", "/home/username/root/file.py", "file"),
            (
                "/home/username/root/",
                "/home/username/root/some/dir/file.py",
                "some.dir.file",
            ),
            # Various special files inside the root should give back valid modules.
            (
                "/home/username/root/",
                "/home/username/root/some/dir/__init__.py",
                "some.dir",
            ),
            (
                "/home/username/root/",
                "/home/username/root/some/dir/__main__.py",
                "some.dir",
            ),
        ),
    )
    def test_calculate_module(
        self, repo_root: Optional[str], filename: str, module: str
    ) -> None:
        self.assertEqual(_calculate_module(repo_root, filename), module)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from typing import Optional

import libcst as cst
from libcst.metadata import BatchableMetadataProvider, MetadataWrapper
from libcst.testing.utils import UnitTest


class MetadataWrapperTest(UnitTest):
    def test_copies_tree(self) -> None:
        m = cst.parse_module("pass")
        mw = MetadataWrapper(m)
        self.assertTrue(mw.module.deep_equals(m))
        self.assertIsNot(mw.module, m)

    def test_unsafe_skip_copy(self) -> None:
        m = cst.parse_module("pass")
        mw = MetadataWrapper(m, unsafe_skip_copy=True)
        self.assertIs(mw.module, m)

    def test_equality_by_identity(self) -> None:
        m = cst.parse_module("pass")
        mw1 = MetadataWrapper(m)
        mw2 = MetadataWrapper(m)
        self.assertEqual(mw1, mw1)
        self.assertEqual(mw2, mw2)
        self.assertNotEqual(mw1, mw2)

    def test_hash_by_identity(self) -> None:
        m = cst.parse_module("pass")
        mw1 = MetadataWrapper(m)
        mw2 = MetadataWrapper(m, unsafe_skip_copy=True)
        mw3 = MetadataWrapper(m, unsafe_skip_copy=True)
        self.assertEqual(hash(mw1), hash(mw1))
        self.assertEqual(hash(mw2), hash(mw2))
        self.assertEqual(hash(mw3), hash(mw3))
        self.assertNotEqual(hash(mw1), hash(mw2))
        self.assertNotEqual(hash(mw1), hash(mw3))
        self.assertNotEqual(hash(mw2), hash(mw3))

    def test_metadata_cache(self) -> None:
        class DummyMetadataProvider(BatchableMetadataProvider[None]):
            gen_cache = tuple

        m = cst.parse_module("pass")
        mw = MetadataWrapper(m)
        with self.assertRaisesRegex(
            Exception, "Cache is required for initializing DummyMetadataProvider."
        ):
            mw.resolve(DummyMetadataProvider)

        class SimpleCacheMetadataProvider(BatchableMetadataProvider[object]):
            gen_cache = tuple

            def __init__(self, cache: object) -> None:
                super().__init__(cache)
                self.cache = cache

            def visit_Pass(self, node: cst.Pass) -> Optional[bool]:
                self.set_metadata(node, self.cache)

        cached_data = object()
        # pyre-fixme[6]: Expected `Mapping[Type[BaseMetadataProvider[object]],
        #  object]` for 2nd param but got `Dict[Type[SimpleCacheMetadataProvider],
        #  object]`.
        mw = MetadataWrapper(m, cache={SimpleCacheMetadataProvider: cached_data})
        pass_node = cst.ensure_type(mw.module.body[0], cst.SimpleStatementLine).body[0]
        self.assertEqual(
            mw.resolve(SimpleCacheMetadataProvider)[pass_node], cached_data
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from types import TracebackType
from typing import Callable, Generator, Iterable, Optional, Type, TypeVar


RetT = TypeVar("RetT")
ArgT = TypeVar("ArgT")


class DummyPool:
    """
    Synchronous dummy `multiprocessing.Pool` analogue.
    """

    def __init__(self, processes: Optional[int] = None) -> None:
        pass

    def imap_unordered(
        self,
        func: Callable[[ArgT], RetT],
        iterable: Iterable[ArgT],
        chunksize: Optional[int] = None,
    ) -> Generator[RetT, None, None]:
        for args in iterable:
            yield func(args)

    def __enter__(self) -> "DummyPool":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[Exception]],
        exc: Optional[Exception],
        tb: Optional[TracebackType],
    ) -> None:
        pass



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from typing import Dict, List, Sequence, Set, Tuple, Union

import libcst
from libcst.codemod._context import CodemodContext
from libcst.codemod._visitor import ContextAwareVisitor
from libcst.helpers import get_absolute_module_for_import


class GatherImportsVisitor(ContextAwareVisitor):
    """
    Gathers all imports in a module and stores them as attributes on the instance.
    Intended to be instantiated and passed to a :class:`~libcst.Module`
    :meth:`~libcst.CSTNode.visit` method in order to gather up information about
    imports on a module. Note that this is not a substitute for scope analysis or
    qualified name support. Please see :ref:`libcst-scope-tutorial` for a more
    robust way of determining the qualified name and definition for an arbitrary
    node.

    After visiting a module the following attributes will be populated:

     module_imports
      A sequence of strings representing modules that were imported directly, such as
      in the case of ``import typing``. Each module directly imported but not aliased
      will be included here.
     object_mapping
      A mapping of strings to sequences of strings representing modules where we
      imported objects from, such as in the case of ``from typing import Optional``.
      Each from import that was not aliased will be included here, where the keys of
      the mapping are the module we are importing from, and the value is a
      sequence of objects we are importing from the module.
     module_aliases
      A mapping of strings representing modules that were imported and aliased,
      such as in the case of ``import typing as t``. Each module imported this
      way will be represented as a key in this mapping, and the value will be
      the local alias of the module.
     alias_mapping
      A mapping of strings to sequences of tuples representing modules where we
      imported objects from and aliased using ``as`` syntax, such as in the case
      of ``from typing import Optional as opt``. Each from import that was aliased
      will be included here, where the keys of the mapping are the module we are
      importing from, and the value is a tuple representing the original object
      name and the alias.
     all_imports
      A collection of all :class:`~libcst.Import` and :class:`~libcst.ImportFrom`
      statements that were encountered in the module.
    """

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)
        # Track the available imports in this transform
        self.module_imports: Set[str] = set()
        self.object_mapping: Dict[str, Set[str]] = {}
        # Track the aliased imports in this transform
        self.module_aliases: Dict[str, str] = {}
        self.alias_mapping: Dict[str, List[Tuple[str, str]]] = {}
        # Track all of the imports found in this transform
        self.all_imports: List[Union[libcst.Import, libcst.ImportFrom]] = []

    def visit_Import(self, node: libcst.Import) -> None:
        # Track this import statement for later analysis.
        self.all_imports.append(node)

        for name in node.names:
            alias = name.evaluated_alias
            if alias is not None:
                # Track this as an aliased module
                self.module_aliases[name.evaluated_name] = alias
            else:
                # Get the module we're importing as a string.
                self.module_imports.add(name.evaluated_name)

    def visit_ImportFrom(self, node: libcst.ImportFrom) -> None:
        # Track this import statement for later analysis.
        self.all_imports.append(node)

        # Get the module we're importing as a string.
        module = get_absolute_module_for_import(self.context.full_module_name, node)
        if module is None:
            # Can't get the absolute import from relative, so we can't
            # support this.
            return
        nodenames = node.names
        if isinstance(nodenames, libcst.ImportStar):
            # We cover everything, no need to bother tracking other things
            self.object_mapping[module] = set("*")
            return
        elif isinstance(nodenames, Sequence):
            # Get the list of imports we're aliasing in this import
            new_aliases = [
                (ia.evaluated_name, ia.evaluated_alias)
                for ia in nodenames
                if ia.asname is not None
            ]
            if new_aliases:
                if module not in self.alias_mapping:
                    self.alias_mapping[module] = []
                # pyre-ignore We know that aliases are not None here.
                self.alias_mapping[module].extend(new_aliases)

            # Get the list of imports we're importing in this import
            new_objects = {ia.evaluated_name for ia in nodenames if ia.asname is None}
            if new_objects:
                if module not in self.object_mapping:
                    self.object_mapping[module] = set()

                # Make sure that we don't add to a '*' module
                if "*" in self.object_mapping[module]:
                    self.object_mapping[module] = set("*")
                    return

                self.object_mapping[module].update(new_objects)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from libcst._batched_visitor import BatchableCSTVisitor, visit_batched
from libcst._exceptions import MetadataException, ParserSyntaxError
from libcst._flatten_sentinel import FlattenSentinel
from libcst._maybe_sentinel import MaybeSentinel
from libcst._metadata_dependent import MetadataDependent
from libcst._nodes.base import CSTNode, CSTValidationError
from libcst._nodes.expression import (
    Annotation,
    Arg,
    Asynchronous,
    Attribute,
    Await,
    BaseAssignTargetExpression,
    BaseComp,
    BaseDelTargetExpression,
    BaseDict,
    BaseDictElement,
    BaseElement,
    BaseExpression,
    BaseFormattedStringContent,
    BaseList,
    BaseNumber,
    BaseSet,
    BaseSimpleComp,
    BaseSlice,
    BaseString,
    BinaryOperation,
    BooleanOperation,
    Call,
    Comparison,
    ComparisonTarget,
    CompFor,
    CompIf,
    ConcatenatedString,
    Dict,
    DictComp,
    DictElement,
    Element,
    Ellipsis,
    Float,
    FormattedString,
    FormattedStringExpression,
    FormattedStringText,
    From,
    GeneratorExp,
    IfExp,
    Imaginary,
    Index,
    Integer,
    Lambda,
    LeftCurlyBrace,
    LeftParen,
    LeftSquareBracket,
    List,
    ListComp,
    Name,
    NamedExpr,
    Param,
    Parameters,
    ParamSlash,
    ParamStar,
    RightCurlyBrace,
    RightParen,
    RightSquareBracket,
    Set,
    SetComp,
    SimpleString,
    Slice,
    StarredDictElement,
    StarredElement,
    Subscript,
    SubscriptElement,
    Tuple,
    UnaryOperation,
    Yield,
)
from libcst._nodes.module import Module
from libcst._nodes.op import (
    Add,
    AddAssign,
    And,
    AssignEqual,
    BaseAugOp,
    BaseBinaryOp,
    BaseBooleanOp,
    BaseCompOp,
    BaseUnaryOp,
    BitAnd,
    BitAndAssign,
    BitInvert,
    BitOr,
    BitOrAssign,
    BitXor,
    BitXorAssign,
    Colon,
    Comma,
    Divide,
    DivideAssign,
    Dot,
    Equal,
    FloorDivide,
    FloorDivideAssign,
    GreaterThan,
    GreaterThanEqual,
    ImportStar,
    In,
    Is,
    IsNot,
    LeftShift,
    LeftShiftAssign,
    LessThan,
    LessThanEqual,
    MatrixMultiply,
    MatrixMultiplyAssign,
    Minus,
    Modulo,
    ModuloAssign,
    Multiply,
    MultiplyAssign,
    Not,
    NotEqual,
    NotIn,
    Or,
    Plus,
    Power,
    PowerAssign,
    RightShift,
    RightShiftAssign,
    Semicolon,
    Subtract,
    SubtractAssign,
)
from libcst._nodes.statement import (
    AnnAssign,
    AsName,
    Assert,
    Assign,
    AssignTarget,
    AugAssign,
    BaseCompoundStatement,
    BaseSmallStatement,
    BaseStatement,
    BaseSuite,
    Break,
    ClassDef,
    Continue,
    Decorator,
    Del,
    Else,
    ExceptHandler,
    Expr,
    Finally,
    For,
    FunctionDef,
    Global,
    If,
    Import,
    ImportAlias,
    ImportFrom,
    IndentedBlock,
    NameItem,
    Nonlocal,
    Pass,
    Raise,
    Return,
    SimpleStatementLine,
    SimpleStatementSuite,
    Try,
    While,
    With,
    WithItem,
)
from libcst._nodes.whitespace import (
    BaseParenthesizableWhitespace,
    Comment,
    EmptyLine,
    Newline,
    ParenthesizedWhitespace,
    SimpleWhitespace,
    TrailingWhitespace,
)
from libcst._parser.entrypoints import parse_expression, parse_module, parse_statement
from libcst._parser.types.config import (
    KNOWN_PYTHON_VERSION_STRINGS,
    PartialParserConfig,
)
from libcst._removal_sentinel import RemovalSentinel, RemoveFromParent
from libcst._version import LIBCST_VERSION
from libcst._visitors import CSTNodeT, CSTTransformer, CSTVisitor, CSTVisitorT
from libcst.helpers import (  # from libcst import ensure_type is deprecated, will be removed in 0.4.0
    ensure_type,
)
from libcst.metadata.base_provider import (
    BaseMetadataProvider,
    BatchableMetadataProvider,
    VisitorMetadataProvider,
)
from libcst.metadata.wrapper import MetadataWrapper


__all__ = [
    "KNOWN_PYTHON_VERSION_STRINGS",
    "LIBCST_VERSION",
    "BatchableCSTVisitor",
    "CSTNodeT",
    "CSTTransformer",
    "CSTValidationError",
    "CSTVisitor",
    "CSTVisitorT",
    "FlattenSentinel",
    "MaybeSentinel",
    "MetadataException",
    "ParserSyntaxError",
    "PartialParserConfig",
    "RemoveFromParent",
    "RemovalSentinel",
    "ensure_type",  # from libcst import ensure_type is deprecated, will be removed in 0.4.0
    "visit_batched",
    "parse_module",
    "parse_expression",
    "parse_statement",
    "CSTNode",
    "Module",
    "Annotation",
    "Arg",
    "Asynchronous",
    "Attribute",
    "Await",
    "BaseAssignTargetExpression",
    "BaseComp",
    "BaseDelTargetExpression",
    "BaseDict",
    "BaseDictElement",
    "BaseElement",
    "BaseExpression",
    "BaseFormattedStringContent",
    "BaseList",
    "BaseNumber",
    "BaseSet",
    "BaseSimpleComp",
    "BaseSlice",
    "BaseString",
    "BinaryOperation",
    "BooleanOperation",
    "Call",
    "Comparison",
    "ComparisonTarget",
    "CompFor",
    "CompIf",
    "ConcatenatedString",
    "Dict",
    "DictComp",
    "DictElement",
    "Element",
    "Ellipsis",
    "Float",
    "FormattedString",
    "FormattedStringExpression",
    "FormattedStringText",
    "From",
    "GeneratorExp",
    "IfExp",
    "Imaginary",
    "Index",
    "Integer",
    "Lambda",
    "LeftCurlyBrace",
    "LeftParen",
    "LeftSquareBracket",
    "List",
    "ListComp",
    "Name",
    "NamedExpr",
    "Param",
    "Parameters",
    "ParamSlash",
    "ParamStar",
    "RightCurlyBrace",
    "RightParen",
    "RightSquareBracket",
    "Set",
    "SetComp",
    "SimpleString",
    "Slice",
    "StarredDictElement",
    "StarredElement",
    "Subscript",
    "SubscriptElement",
    "Tuple",
    "UnaryOperation",
    "Yield",
    "Add",
    "AddAssign",
    "And",
    "AssignEqual",
    "BaseAugOp",
    "BaseBinaryOp",
    "BaseBooleanOp",
    "BaseCompOp",
    "BaseUnaryOp",
    "BitAnd",
    "BitAndAssign",
    "BitInvert",
    "BitOr",
    "BitOrAssign",
    "BitXor",
    "BitXorAssign",
    "Colon",
    "Comma",
    "Divide",
    "DivideAssign",
    "Dot",
    "Equal",
    "FloorDivide",
    "FloorDivideAssign",
    "GreaterThan",
    "GreaterThanEqual",
    "ImportStar",
    "In",
    "Is",
    "IsNot",
    "LeftShift",
    "LeftShiftAssign",
    "LessThan",
    "LessThanEqual",
    "MatrixMultiply",
    "MatrixMultiplyAssign",
    "Minus",
    "Modulo",
    "ModuloAssign",
    "Multiply",
    "MultiplyAssign",
    "Not",
    "NotEqual",
    "NotIn",
    "Or",
    "Plus",
    "Power",
    "PowerAssign",
    "RightShift",
    "RightShiftAssign",
    "Semicolon",
    "Subtract",
    "SubtractAssign",
    "AnnAssign",
    "AsName",
    "Assert",
    "Assign",
    "AssignTarget",
    "AugAssign",
    "BaseCompoundStatement",
    "BaseSmallStatement",
    "BaseStatement",
    "BaseSuite",
    "Break",
    "ClassDef",
    "Continue",
    "Decorator",
    "Del",
    "Else",
    "ExceptHandler",
    "Expr",
    "Finally",
    "For",
    "FunctionDef",
    "Global",
    "If",
    "Import",
    "ImportAlias",
    "ImportFrom",
    "IndentedBlock",
    "NameItem",
    "Nonlocal",
    "Pass",
    "Raise",
    "Return",
    "SimpleStatementLine",
    "SimpleStatementSuite",
    "Try",
    "While",
    "With",
    "WithItem",
    "BaseParenthesizableWhitespace",
    "Comment",
    "EmptyLine",
    "Newline",
    "ParenthesizedWhitespace",
    "SimpleWhitespace",
    "TrailingWhitespace",
    "BaseMetadataProvider",
    "BatchableMetadataProvider",
    "VisitorMetadataProvider",
    "MetadataDependent",
    "MetadataWrapper",
]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst import parse_statement
from libcst._nodes.tests.base import CSTNodeTest, DummyIndentedBlock
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class SimpleStatementTest(CSTNodeTest):
    @data_provider(
        (
            # a single-element SimpleStatementLine
            {
                "node": cst.SimpleStatementLine((cst.Pass(),)),
                "code": "pass\n",
                "parser": parse_statement,
            },
            # a multi-element SimpleStatementLine
            {
                "node": cst.SimpleStatementLine(
                    (cst.Pass(semicolon=cst.Semicolon()), cst.Continue())
                ),
                "code": "pass;continue\n",
                "parser": parse_statement,
            },
            # a multi-element SimpleStatementLine with whitespace
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Pass(
                            semicolon=cst.Semicolon(
                                whitespace_before=cst.SimpleWhitespace(" "),
                                whitespace_after=cst.SimpleWhitespace("  "),
                            )
                        ),
                        cst.Continue(),
                    )
                ),
                "code": "pass ;  continue\n",
                "parser": parse_statement,
            },
            # A more complicated SimpleStatementLine
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Pass(semicolon=cst.Semicolon()),
                        cst.Continue(semicolon=cst.Semicolon()),
                        cst.Break(),
                    )
                ),
                "code": "pass;continue;break\n",
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (1, 19)),
            },
            # a multi-element SimpleStatementLine, inferred semicolons
            {
                "node": cst.SimpleStatementLine(
                    (cst.Pass(), cst.Continue(), cst.Break())
                ),
                "code": "pass; continue; break\n",
                "parser": None,  # No test for parsing, since we are using sentinels.
            },
            # some expression statements
            {
                "node": cst.SimpleStatementLine((cst.Expr(cst.Name("None")),)),
                "code": "None\n",
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine((cst.Expr(cst.Name("True")),)),
                "code": "True\n",
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine((cst.Expr(cst.Name("False")),)),
                "code": "False\n",
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine((cst.Expr(cst.Ellipsis()),)),
                "code": "...\n",
                "parser": parse_statement,
            },
            # Test some numbers
            {
                "node": cst.SimpleStatementLine((cst.Expr(cst.Integer("5")),)),
                "code": "5\n",
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine((cst.Expr(cst.Float("5.5")),)),
                "code": "5.5\n",
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine((cst.Expr(cst.Imaginary("5j")),)),
                "code": "5j\n",
                "parser": parse_statement,
            },
            # Test some numbers with parens
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Expr(
                            cst.Integer(
                                "5", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                            )
                        ),
                    )
                ),
                "code": "(5)\n",
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (1, 3)),
            },
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Expr(
                            cst.Float(
                                "5.5", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                            )
                        ),
                    )
                ),
                "code": "(5.5)\n",
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Expr(
                            cst.Imaginary(
                                "5j", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                            )
                        ),
                    )
                ),
                "code": "(5j)\n",
                "parser": parse_statement,
            },
            # Test some strings
            {
                "node": cst.SimpleStatementLine((cst.Expr(cst.SimpleString('"abc"')),)),
                "code": '"abc"\n',
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Expr(
                            cst.ConcatenatedString(
                                cst.SimpleString('"abc"'), cst.SimpleString('"def"')
                            )
                        ),
                    )
                ),
                "code": '"abc""def"\n',
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Expr(
                            cst.ConcatenatedString(
                                left=cst.SimpleString('"abc"'),
                                whitespace_between=cst.SimpleWhitespace(" "),
                                right=cst.ConcatenatedString(
                                    left=cst.SimpleString('"def"'),
                                    whitespace_between=cst.SimpleWhitespace(" "),
                                    right=cst.SimpleString('"ghi"'),
                                ),
                            )
                        ),
                    )
                ),
                "code": '"abc" "def" "ghi"\n',
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (1, 17)),
            },
            # Test parenthesis rules
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Expr(
                            cst.Ellipsis(
                                lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                            )
                        ),
                    )
                ),
                "code": "(...)\n",
                "parser": parse_statement,
            },
            # Test parenthesis with whitespace ownership
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Expr(
                            cst.Ellipsis(
                                lpar=(
                                    cst.LeftParen(
                                        whitespace_after=cst.SimpleWhitespace(" ")
                                    ),
                                ),
                                rpar=(
                                    cst.RightParen(
                                        whitespace_before=cst.SimpleWhitespace(" ")
                                    ),
                                ),
                            )
                        ),
                    )
                ),
                "code": "( ... )\n",
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Expr(
                            cst.Ellipsis(
                                lpar=(
                                    cst.LeftParen(
                                        whitespace_after=cst.SimpleWhitespace(" ")
                                    ),
                                    cst.LeftParen(
                                        whitespace_after=cst.SimpleWhitespace("  ")
                                    ),
                                    cst.LeftParen(
                                        whitespace_after=cst.SimpleWhitespace("   ")
                                    ),
                                ),
                                rpar=(
                                    cst.RightParen(
                                        whitespace_before=cst.SimpleWhitespace("   ")
                                    ),
                                    cst.RightParen(
                                        whitespace_before=cst.SimpleWhitespace("  ")
                                    ),
                                    cst.RightParen(
                                        whitespace_before=cst.SimpleWhitespace(" ")
                                    ),
                                ),
                            )
                        ),
                    )
                ),
                "code": "( (  (   ...   )  ) )\n",
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (1, 21)),
            },
            # Test parenthesis rules with expressions
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Expr(
                            cst.Ellipsis(
                                lpar=(
                                    cst.LeftParen(
                                        whitespace_after=cst.ParenthesizedWhitespace(
                                            first_line=cst.TrailingWhitespace(),
                                            empty_lines=(
                                                cst.EmptyLine(
                                                    comment=cst.Comment(
                                                        "# Wow, a comment!"
                                                    )
                                                ),
                                            ),
                                            indent=True,
                                            last_line=cst.SimpleWhitespace("    "),
                                        )
                                    ),
                                ),
                                rpar=(
                                    cst.RightParen(
                                        whitespace_before=cst.ParenthesizedWhitespace(
                                            first_line=cst.TrailingWhitespace(),
                                            empty_lines=(),
                                            indent=True,
                                            last_line=cst.SimpleWhitespace(""),
                                        )
                                    ),
                                ),
                            )
                        ),
                    )
                ),
                "code": "(\n# Wow, a comment!\n    ...\n)\n",
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (4, 1)),
            },
            # test trailing whitespace
            {
                "node": cst.SimpleStatementLine(
                    (cst.Pass(),),
                    trailing_whitespace=cst.TrailingWhitespace(
                        whitespace=cst.SimpleWhitespace("  "),
                        comment=cst.Comment("# trailing comment"),
                    ),
                ),
                "code": "pass  # trailing comment\n",
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (1, 4)),
            },
            # test leading comment
            {
                "node": cst.SimpleStatementLine(
                    (cst.Pass(),),
                    leading_lines=(cst.EmptyLine(comment=cst.Comment("# comment")),),
                ),
                "code": "# comment\npass\n",
                "parser": parse_statement,
                "expected_position": CodeRange((2, 0), (2, 4)),
            },
            # test indentation
            {
                "node": DummyIndentedBlock(
                    "    ",
                    cst.SimpleStatementLine(
                        (cst.Pass(),),
                        leading_lines=(
                            cst.EmptyLine(comment=cst.Comment("# comment")),
                        ),
                    ),
                ),
                "code": "    # comment\n    pass\n",
                "expected_position": CodeRange((2, 4), (2, 8)),
            },
            # test suite variant
            {
                "node": cst.SimpleStatementSuite((cst.Pass(),)),
                "code": " pass\n",
                "expected_position": CodeRange((1, 1), (1, 5)),
            },
            {
                "node": cst.SimpleStatementSuite(
                    (cst.Pass(),), leading_whitespace=cst.SimpleWhitespace("")
                ),
                "code": "pass\n",
                "expected_position": CodeRange((1, 0), (1, 4)),
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from collections import Counter
from textwrap import dedent

import libcst as cst
from libcst.testing.utils import UnitTest, data_provider


class DuplicateLeafNodeTest(UnitTest):
    @data_provider(
        (
            # Simple program
            (
                """
                foo = 'toplevel'
                fn1(foo)
                fn2(foo)
                def fn_def():
                    foo = 'shadow'
                    fn3(foo)
            """,
            ),
        )
    )
    def test_tokenize(self, code: str) -> None:
        test_case = self

        class CountVisitor(cst.CSTVisitor):
            def __init__(self) -> None:
                self.count = Counter()
                self.nodes = {}

            def on_visit(self, node: cst.CSTNode) -> bool:
                self.count[id(node)] += 1
                test_case.assertTrue(
                    self.count[id(node)] == 1,
                    f"Node duplication detected between {node} and {self.nodes.get(id(node))}",
                )
                self.nodes[id(node)] = node
                return True

        module = cst.parse_module(dedent(code))
        module.visit(CountVisitor())
